#pragma once

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <curl/curl.h>
#include <json-c/json.h>

#include "custom_log.h"

/* struct where curl responses are stored */
struct MemoryStruct {
  char *memory;
  size_t size;
};

double diff_time(struct timespec* begin, struct timespec* end);

void stopwatch_start(struct timespec* w_begin, struct timespec* cpu_begin);

void stopwatch_stop(struct timespec* w_begin, struct timespec* w_end, struct timespec* cpu_begin, struct timespec* cpu_end, 
    char* message, double* w_total, double* cpu_total);

/* It gets response from Key Server and sanitizes it removing spaces and chars related to encoding format */
void sanitize_ks_response(char*);

/* Callback called from curl to store data from HTTP response */
size_t write_memory_cb(void*, size_t, size_t, void*);

/* Send POST request to keycloak to obtain access token */
int get_token(struct MemoryStruct* chunk, const char* keycloak_ip, const char* keycloak_port, const char* keycloak_client_id, const char* keycloak_sec);

/* Send POST request to KeyServer to reserve one key */
int get_key(struct MemoryStruct* chunk, const char* token, const char* key_server_ip, const char* key_server_port, const char* target_sae, const char* this_sae_str);

/* Send POST request to KeyServer to retrieve one key sending the related key_id */
int get_key_with_id(struct MemoryStruct*, const char* token, const char* key_id, const char* key_server_ip, const char* key_server_port, const char* target_sae);

/* 
Parse token from JSON HTTP response from Keycloak
INPUT:
    - char* buffer     -> string to be parsed
    - char** token_str -> pointer to memory area where token will be stored
OUTPUT:
    - 0 if executed correctly
    - 1 otherwise
*/
int parse_json_token(char*, char**);

/* Parse key and id from JSON HTTP response from KeyServer after get_key() */
int parse_json_key(char*, char**, char**);

/* Parse key and id from JSON HTTP response from KeyServer after get_key_with_id() */
int parse_json_key_id(char*, char**, char**);

