import os
import signal
import socket
import requests
import json
import subprocess
import logging
import configparser
import time
from subprocess import check_output
from ast import literal_eval
from threading import Thread
from threading import Lock
from queue import Queue

keyFileLock = Lock()

N_THREAD = 20

def getToken(kcIP, kcPort, kcRealm, kcSAE, kcCredentials):
    x = requests.post('http://' + kcIP + ':' + kcPort + '/auth/realms/' + kcRealm + '/protocol/openid-connect/token', data='client_id=' + kcSAE + '&client_secret=' + kcCredentials + '&grant_type=client_credentials', headers={'Content-Type':'application/x-www-form-urlencoded'})
    token = x.json()['access_token']
    oauthHeader = {'Authorization': 'Bearer ' + token}
    return oauthHeader

def getFromKeyServer(keyServerIP, keyServerPort, keyID, targetSAE, thisSAE, kNum, kLen, oauthHeader):

    if(keyID == -1):  # sender side
        x = requests.post('http://' + keyServerIP + ':' + keyServerPort + '/api/v1/keys/' + targetSAE + '/enc_keys', data=json.dumps({'number' : kNum, 'size' : kLen, 'SAE_ID' : thisSAE}), headers=oauthHeader)
        if x.status_code != 200:
            print("Get key failed with following error code:", x.status_code, x.content)
            return

        result = x.content
        print("\nGET KEY result:\n", result)
        res = x.json()
        res = literal_eval(res)
        key = res[0]['key'].decode()
        keyID = res[0]['key_ID']
        keyFileLock.acquire()
        try: 
            keyFile.write(f"{keyID}:{key}\n")
        finally:
            keyFileLock.release()
        return keyID

    else:   # receiver side
        kIDs = []
        kIDs.append({'key_ID' : keyID})
        x = requests.post('http://' + keyServerIP + ':' + keyServerPort + '/api/v1/keys/' + targetSAE + '/dec_keys', data=json.dumps({'key_IDs' : kIDs}), headers=oauthHeader)
        if x.status_code != 200:
            print("Get key with key ID failed with following error code:", x.status_code, x.content)
            return
 
        result = x.content
        print("\nGET KEY WITH KEY ID result:\n", result)
        res = x.json()
        res = res['keys']
        res = literal_eval(res)
        key = res[0]['key'].decode()
        return key

def threadFunc(config, id):

    startThreadTime = time.perf_counter()

    confKeycloak = config['Keycloak']
    confKeyserver = config['KeyServer']
    confConn = config['Connection']

    tokenStartTime = time.perf_counter()
    oauthHeader = getToken(confKeycloak['keycloakIP'], confKeycloak['keycloakPort'], confKeycloak['keycloakRealm'], confKeycloak['clientSaeID'], confKeycloak['clientCredentials'])
    tokenTime = time.perf_counter() - tokenStartTime

    getKeyStartTime = time.perf_counter()
    keyID = getFromKeyServer(confKeyserver['keyServIP'], confKeyserver['keyServPort'], -1, confConn['targetSAE'], confConn['thisSAE'], confConn['kNum'], confConn['kLen'], oauthHeader)
    getKeyTime = time.perf_counter() - getKeyStartTime

    sendHandleStartTime = time.perf_counter()
    while True:
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect((confConn['targetIP'], int(confConn['targetPort'])))
            break
        except socket.error:
            print(f'[THREAD-{id}]: Connection refuse, retrying...')
            time.sleep(1)
    s.sendall(('SET KEY ID=' + keyID + '\r\n\r\n').encode('utf-8'))
    sendHandleTime = time.perf_counter() - sendHandleStartTime

    recvACKstartTime = time.perf_counter()
    recvData = bytearray()
    while True:
        packet = s.recv(4096)
        recvData.extend(packet)
        if(recvData.decode('utf-8').endswith('\r\n') or recvData.decode('utf-8').endswith('\n')):
            break
    message = recvData.decode('utf-8')
    s.close()
    recvACKtime = time.perf_counter() - recvACKstartTime

    if(message.startswith("OK DONE!")):
        threadTime = time.perf_counter() - startThreadTime
        logstr = f"[THREAD-{id}]: get token = {tokenTime:0.4f} s, get key = {getKeyTime:0.4f} s, send handle = {sendHandleTime:0.4f} s, recv ACK = {recvACKtime:0.4f} s, Total exec time: {threadTime:0.4f} s"
        logging.info('%s', logstr)

    

def main():
    logging.basicConfig(filename="multiGet.log", filemode='w', level=logging.INFO, format='%(levelname)s %(asctime)s %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p')
    config = configparser.ConfigParser()
    config.read('config.ini')

    global keyFile
    keyFile = open('client_keys', 'w')

    threads = []

    for i in range(N_THREAD):
        worker = Thread(target=threadFunc, args=(config, i, ))
        threads.append(worker)
        worker.start()

    for t in threads:
        t.join()




if __name__ == '__main__':
    main()