import os
import signal
import socket
import requests
import json
import subprocess
import logging
import configparser
import time
from subprocess import check_output
from ast import literal_eval
from threading import Thread
from threading import Lock
from queue import Queue

keyFileLock = Lock()

MSG_REQ = "SET KEY ID="

N_THREAD = 20

def getToken(kcIP, kcPort, kcRealm, kcSAE, kcCredentials):
    x = requests.post('http://' + kcIP + ':' + kcPort + '/auth/realms/' + kcRealm + '/protocol/openid-connect/token', data='client_id=' + kcSAE + '&client_secret=' + kcCredentials + '&grant_type=client_credentials', headers={'Content-Type':'application/x-www-form-urlencoded'})
    token = x.json()['access_token']
    oauthHeader = {'Authorization': 'Bearer ' + token}
    return oauthHeader

def getFromKeyServer(keyServerIP, keyServerPort, keyID, targetSAE, thisSAE, kNum, kLen, oauthHeader):

    if(keyID == -1):  # sender side
        x = requests.post('http://' + keyServerIP + ':' + keyServerPort + '/api/v1/keys/' + targetSAE + '/enc_keys', data=json.dumps({'number' : kNum, 'size' : kLen, 'SAE_ID' : thisSAE}), headers=oauthHeader)
        if x.status_code != 200:
            print("Get key failed with following error code:", x.status_code, x.content)
            return
        
        result = x.content
        print("\nGET KEY result:\n", result)
        res = x.json()
        res = literal_eval(res)
        key = res[0]['key'].decode()
        keyID = res[0]['key_ID']
        stunnPskLock.acquire()
        try: 
            stunnPsk.addOnlyLast(keyID, key)
        finally:
            stunnPskLock.release()
        return keyID

    else:   # receiver side
        kIDs = []
        kIDs.append({'key_ID' : keyID})
        x = requests.post('http://' + keyServerIP + ':' + keyServerPort + '/api/v1/keys/' + targetSAE + '/dec_keys', data=json.dumps({'key_IDs' : kIDs}), headers=oauthHeader)
        if x.status_code != 200:
            print("Get key with key ID failed with following error code:", x.status_code, x.content)
            return
        result = x.content
        print("\nGET KEY WITH KEY ID result:\n", result)
        res = x.json()
        res = res['keys']
        res = literal_eval(res)
        key = res[0]['key'].decode()
        return key

def threadFunc(config, id, c):

    startThreadTime = time.perf_counter()

    confKeycloak = config['Keycloak']
    confKeyserver = config['KeyServer']
    confConn = config['Connection']

    tokenStartTime = time.perf_counter()
    oauthHeader = getToken(confKeycloak['keycloakIP'], confKeycloak['keycloakPort'], confKeycloak['keycloakRealm'], confKeycloak['clientSaeID'], confKeycloak['clientCredentials'])
    tokenTime = time.perf_counter() - tokenStartTime

    recvHandlestartTime = time.perf_counter()
    recvData = bytearray()
    while True:
        packet = c.recv(4096)
        recvData.extend(packet)
        if(recvData.decode('utf-8').endswith('\r\n') or recvData.decode('utf-8').endswith('\n')):
            break
    message = recvData.decode('utf-8')
    if(message.startswith(MSG_REQ)):
        # correct format message
        keyID = message.split("=")[1].rstrip()
    else:
        reply = "MESSAGE FORMAT ERROR\r\n" 
        c.send(reply.encode())
        c.close()
        return
    recvHandleTime = time.perf_counter() - recvHandlestartTime

    getKeyStartTime = time.perf_counter()
    key = getFromKeyServer(confKeyserver['keyServIP'], confKeyserver['keyServPort'], keyID, confConn['targetSAE'], confConn['thisSAE'], confConn['kNum'], confConn['kLen'], oauthHeader)
    keyFileLock.acquire()
    try: 
        keyFile.write(f"{keyID}:{key}\n")
    finally:
        keyFileLock.release()
    getKeyTime = time.perf_counter() - getKeyStartTime

    sendACKStartTime = time.perf_counter()
    reply = "OK DONE!\r\n" 
    c.send(reply.encode())
    c.close()
    sendACKTime = time.perf_counter() - sendACKStartTime

    threadTime = time.perf_counter() - startThreadTime
    logstr = f"[THREAD-{id}]: get token = {tokenTime:0.4f} s, recv handle = {recvHandleTime:0.4f} s, get key with id = {getKeyTime:0.4f} s, send ACK = {sendACKTime:0.4f} s, Total exec time: {threadTime:0.4f} s"
    logging.info('%s', logstr)

    

def main():
    logging.basicConfig(filename="multiGet.log", filemode='w', level=logging.INFO, format='%(levelname)s %(asctime)s %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p')
    config = configparser.ConfigParser()
    config.read('config.ini')

    global keyFile
    keyFile = open('server_keys', 'w')

    threads = []

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM) 
    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    s.bind((config['Server']['IPaddr'], int(config['Server']['port']))) 
    s.listen(5) 
  
    for i in range(N_THREAD): 

        c, addr = s.accept() 
        # Start a new thread and return its identifier 
        t = Thread(target=threadFunc, args=(config, i, c, ))
        t.start()
        threads.append(t)

    for t in threads:
        t.join()



if __name__ == '__main__':
    main()