import socket
from queue import Queue
from threading import Thread
from threading import Lock
from Timer import Timer

MSG_REQ = "SET KEY ID="

##################################################
# thread function 
def readData(c, host, port, q: Queue, timer: Timer):
    
    data = bytearray()

    while True:
        packet = c.recv(1024)
        data.extend(packet)
        if(data.decode('utf-8').endswith('\r\n') or data.decode('utf-8').endswith('\n')):
            break

    message = data.decode('utf-8')
    if(message.startswith(MSG_REQ)):
        # correct format message
        keyID = message.split("=")[1].rstrip()
        msg = (c, host, port, keyID)
        timer.stop("Thread created, Handle retrieved from client")
        q.put(msg)
    else:
        reply = "MESSAGE FORMAT ERROR\r\n" 
        c.send(reply.encode())
        c.close()

##################################################


##################################################
class PlainServer:
    
    ##################################################
    def __init__(self, host, port, q: Queue, timer: Timer):
        self.host = host
        self.port = port
        self.q = q
        self.timer = timer
        self.threads = []
    ##################################################

    ##################################################
    def startLoop(self):
        self.timer.start()
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM) 
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.bind((self.host, self.port)) 
        print("socket binded to port", self.port) 
    
        # put the socket into listening mode 
        s.listen(5) 
        print("socket is listening") 
    
        # a forever loop until client wants to exit 
        while True: 
    
            self.timer.stop("Plain text server ready to accept connections")
            # establish connection with client 
            c, addr = s.accept() 
            self.timer.start()
            # lock acquired by client
            print('Connected to :', addr[0], ':', addr[1]) 
    
            # Start a new thread and return its identifier 
            t = Thread(target= readData, args= (c, addr[0], addr[1], self.q, self.timer))
            t.start()
            self.threads.append(t)

            # for debug and test reason exits after first connection
            break
        
        for thread in self.threads:
            thread.join()
        s.close()
    ##################################################

##################################################
     