import time
import logging

class TimerError(Exception):
    """A custom exception used to report errors in use of Timer class"""

class Timer:

    tot_time = float(0)

    def __init__(self, procTypeMSG):
        self._start_time = None
        self.procTypeMSG = procTypeMSG

    def start(self):
        """Start a new timer"""
        if self._start_time is not None:
            raise TimerError(f"Timer is running. Use .stop() to stop it")
        self._start_time = time.perf_counter()

    def stop(self, msg, end=False):
        """Stop the timer, and report the elapsed time"""
        if self._start_time is None:
            raise TimerError(f"Timer is not running. Use .start() to start it")
        elapsed_time = time.perf_counter() - self._start_time
        self._start_time = None
        logstr = f"Elapsed time: {elapsed_time:0.4f} seconds"
        if (msg):
            logstr = self.procTypeMSG + ' ' + msg + ' ' + logstr
        logging.info('%s', logstr)
        self.tot_time += elapsed_time
        if (end):
            logstr = f"Total time: {self.tot_time:0.4f} seconds\n"
            logging.info('%s', self.procTypeMSG + ' ' + logstr)
            