import os
import signal
import socket
import requests
import json
import subprocess
import psutil
import logging
import sys
from threading import Thread
from subprocess import check_output
from ast import literal_eval
from ManageSecrets import ManageSecrets
from Timer import Timer

procTypeMSG = "[STANDARD-PSK]"

configs = {}            # Dictionary used to store configs read from bootConn.conf

##################################################
# Used to check if stunnel is already running
##################################################

def checkIfProcessRunning(processName):
    '''
    Check if there is any running process that contains the given name processName.
    '''
    #Iterate over the all the running process
    for proc in psutil.process_iter():
        try:
            # Check if process name contains the given name string.
            if processName.lower() in proc.name().lower():
                return True
        except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
            pass
    return False
 
##################################################



##################################################
# Used to retrieve stunnel's pid
##################################################

def findProcessIdByName(processName):
    '''
    Get a list of all the PIDs of a all the running process whose name contains
    the given string processName
    '''
    listOfProcessObjects = []
    #Iterate over the all the running process
    for proc in psutil.process_iter():
       try:
           pinfo = proc.as_dict(attrs=['pid', 'name', 'create_time'])
           # Check if process name contains the given name string.
           if processName.lower() in pinfo['name'].lower() :
               listOfProcessObjects.append(pinfo)
       except (psutil.NoSuchProcess, psutil.AccessDenied , psutil.ZombieProcess) :
           pass
    return listOfProcessObjects

##################################################



##################################################
# launch stunnel process storing its reference
##################################################

def startStunnel():
    execStun = subprocess.Popen(["sudo", "stunnel"])
    execStun.wait()

##################################################



if __name__ == '__main__':
    
    logging.basicConfig(filename="setupStunnPSK.log", level=logging.INFO, format='%(levelname)s %(asctime)s %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p')
    global timer
    timer = Timer(procTypeMSG)

    timer.start()
    # get configs from bootConn.conf
    with open('./bootConn.conf', 'r') as f:
        data = f.readlines()
        for line in data:
            if (line.startswith("\n")):
                continue
            key, val = line.split("=")
            if (val.find("#") != -1):
                val, _ = val.split("#")
            configs[key.rstrip()] = val.replace(" ", "").rstrip()
    timer.stop("Loaded configuration options from file")

    timer.start()
    # file to store ID:PSK pairs for stunnel exchanges
    global stunnPsk 
    stunnPsk = ManageSecrets(configs.get('stunnPsk'))
    timer.stop("Created/Loaded secrets file")

    timer.start()
    if (sys.argv[1]):
        keyID = sys.argv[1]
        key = sys.argv[2]
    else:
        exit
    stunnPsk.addOnlyLast(keyID, key)
    timer.stop("Key parsed and stored on file")

    timer.start()
    if (checkIfProcessRunning("stunnel")):
        lPids = findProcessIdByName("stunnel")
        stunnPid = lPids[0]['pid']
        reconfProc = subprocess.Popen(["sudo", "kill", "-SIGHUP", str(stunnPid)])
        reconfProc.wait()
        print('Stunnel correctly reconfigured')
        timer.stop("Stunnel reconfigured", True)
    else:
        stunnelThread = Thread(target = startStunnel,)
        stunnelThread.start()
        timer.stop("Stunnel started", True)

