import os
import signal
import socket
import requests
import json
import subprocess
import psutil
import logging
from subprocess import check_output
from ast import literal_eval
from PlainTextServer import PlainServer
from threading import Thread
from threading import Lock
from queue import Queue
from ManageSecrets import ManageSecrets
from Timer import Timer

procTypeMSG = "[PSK-KEY-SERVER]"

configs = {}            # Dictionary used to store configs read from bootConn.conf

qPlain = Queue()        # Queue used to get messages from clients --- PlainText

stunnPskLock = Lock()

isStunnelActive = 0

##################################################
# Used to check if stunnel is already running
##################################################

def checkIfProcessRunning(processName):
    '''
    Check if there is any running process that contains the given name processName.
    '''
    #Iterate over the all the running process
    for proc in psutil.process_iter():
        try:
            # Check if process name contains the given name string.
            if processName.lower() in proc.name().lower():
                return True
        except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
            pass
    return False
 
##################################################



##################################################
# Used to retrieve stunnel's pid
##################################################

def findProcessIdByName(processName):
    '''
    Get a list of all the PIDs of a all the running process whose name contains
    the given string processName
    '''
    listOfProcessObjects = []
    #Iterate over the all the running process
    for proc in psutil.process_iter():
       try:
           pinfo = proc.as_dict(attrs=['pid', 'name', 'create_time'])
           # Check if process name contains the given name string.
           if processName.lower() in pinfo['name'].lower() :
               listOfProcessObjects.append(pinfo)
       except (psutil.NoSuchProcess, psutil.AccessDenied , psutil.ZombieProcess) :
           pass
    return listOfProcessObjects

##################################################



##################################################
# Function to retrieve token from Keycloak
##################################################

def getToken(kcIP, kcPort, kcRealm, kcSAE, kcCredentials):
    x = requests.post('http://' + kcIP + ':' + kcPort + '/auth/realms/' + kcRealm + '/protocol/openid-connect/token', data='client_id=' + kcSAE + '&client_secret=' + kcCredentials + '&grant_type=client_credentials', headers={'Content-Type':'application/x-www-form-urlencoded'})
    token = x.json()['access_token']
    oauthHeader = {'Authorization': 'Bearer ' + token}
    return oauthHeader

##################################################
 


##################################################
# Function to retrieve key with ID from Key Server
##################################################

def getFromKeyServer(keyServerIP, keyServerPort, keyID, targetSAE, thisSAE, kNum, kLen, oauthHeader):

    if(keyID == -1):  # sender side
        timer.start()
        x = requests.post('http://' + keyServerIP + ':' + keyServerPort + '/api/v1/keys/' + targetSAE + '/enc_keys', data=json.dumps({'number' : kNum, 'size' : kLen, 'SAE_ID' : thisSAE}), headers=oauthHeader)
        if x.status_code != 200:
            print("Get key failed with following error code:", x.status_code, x.content)
            return
        timer.stop("Received key from KeyServer (GET KEY)")

        timer.start()
        result = x.content
        print("\nGET KEY result:\n", result)
        res = x.json()
        res = literal_eval(res)
        key = res[0]['key'].decode()
        keyID = res[0]['key_ID']
        stunnPskLock.acquire()
        try: 
            stunnPsk.addOnlyLast(keyID, key)
        finally:
            stunnPskLock.release()
        timer.stop("Key parsed and stored on file")
        return keyID

    else:   # receiver side
        kIDs = []
        kIDs.append({'key_ID' : keyID})
        timer.start()
        x = requests.post('http://' + keyServerIP + ':' + keyServerPort + '/api/v1/keys/' + targetSAE + '/dec_keys', data=json.dumps({'key_IDs' : kIDs}), headers=oauthHeader)
        if x.status_code != 200:
            print("Get key with key ID failed with following error code:", x.status_code, x.content)
            return
        timer.stop("Received key from KeyServer (GET KEY WITH ID)")
        timer.start()
        result = x.content
        print("\nGET KEY WITH KEY ID result:\n", result)
        res = x.json()
        res = res['keys']
        res = literal_eval(res)
        key = res[0]['key'].decode()
        return key

##################################################


##################################################
# Boot Server to listen to clients requests
##################################################

def BootPlainServer(host, port): 
    s = PlainServer(host, port, qPlain, timer)
    s.startLoop()

##################################################


##################################################
# Infinite Loop to respond to each client request
# Dummy version uses the received key to set stunnel
##################################################

def plainConsumerLoopDummy(keyServerIP, keyServerPort, oauthHeader):
    
    while True:

        msg = qPlain.get()
        c, address, port, keyID = msg

        # get key from Key Server for key ID
        key = getFromKeyServer(keyServerIP, keyServerPort, keyID, configs.get('targetSAE'), configs.get('serverSAE'), configs.get('kNum'), configs.get('kLen'), oauthHeader)

        # edit secretsFp (dummy version, in actual version should store the key for tls-psk server)
        stunnPskLock.acquire()
        try: 
            stunnPsk.addOnlyLast(keyID, key)
        finally:
            stunnPskLock.release()
        timer.stop("Key parsed and stored on file")

        timer.start()
        if (checkIfProcessRunning("stunnel")):
            lPids = findProcessIdByName("stunnel")
            stunnPid = lPids[0]['pid']
            reconfProc = subprocess.Popen(["sudo", "kill", "-SIGHUP", str(stunnPid)])
            reconfProc.wait()
            print('Stunnel correctly reconfigured')
            timer.stop("Stunnel reconfigured")
        else:
            stunnelThread = Thread(target = startStunnel,)
            stunnelThread.start()
            timer.stop("Stunnel started")

        timer.start()
        # reply to client using c and close connection
        reply = "OK DONE!\r\n" 
        c.send(reply.encode())
        c.close()
        timer.stop("Sent ACK to client", True)

        # to debug and test only one connection configured then exits
        break

##################################################



##################################################
# launch stunnel process storing its reference
##################################################

def startStunnel():
    execStun = subprocess.Popen(["sudo", "stunnel"])
    execStun.wait()

##################################################



if __name__ == '__main__':
    
    logging.basicConfig(filename="setupStunn.log", level=logging.INFO, format='%(levelname)s %(asctime)s %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p')
    global timer
    timer = Timer(procTypeMSG)

    timer.start()
    # get configs from bootConn.conf
    with open('./bootConn.conf', 'r') as f:
        data = f.readlines()
        for line in data:
            if (line.startswith("\n")):
                continue
            key, val = line.split("=")
            if (val.find("#") != -1):
                val, _ = val.split("#")
            configs[key.rstrip()] = val.replace(" ", "").rstrip()
    timer.stop("Loaded configuration options from file")

    timer.start()
    # file to store ID:PSK pairs for stunnel exchanges
    global stunnPsk 
    stunnPsk = ManageSecrets(configs.get('stunnPsk'))
    timer.stop("Created/Loaded secrets file")

    timer.start()
    # Get token from keycloak
    oauthHeader = getToken(configs.get('keycloakIP'), configs.get('keycloakPort'), configs.get('keycloakRealm'), configs.get('clientSaeID'), configs.get('clientCredentials'))
    timer.stop("Retrieved token from keycloak")

    # Create and start PlainTextServer thread
    # for debug and testing sender won't create it
    if (int(configs.get('sender')) != 1):
        plainServerThread = Thread(target = BootPlainServer, args = (configs.get('IPaddr'), int(configs.get('port')), ))
        plainServerThread.start()

    # Create and start PlainTextConsumer thread
    # for debug and testing sender won't create it
    if (int(configs.get('sender')) != 1):
        plainConsumer = Thread(target = plainConsumerLoopDummy, args = (configs.get('keyServIP'), configs.get('keyServPort'), oauthHeader, ))
        plainConsumer.start()

    # if sender == 1 then get key and send handle to peer
    if (int(configs.get('sender'))):
        keyID = getFromKeyServer(configs.get("keyServIP"), configs.get("keyServPort"), -1, configs.get('targetSAE'), configs.get('clientSaeID'), int(configs.get('kNum')), int(configs.get('kLen')), oauthHeader)
        
        timer.start()
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)   # create TCP socket
        s.connect((configs.get('targetIP'), int(configs.get('targetPort'))))
        s.sendall(('SET KEY ID=' + keyID + '\r\n\r\n').encode('utf-8'))
        timer.stop("Handle sent to server")

        recvData = bytearray()
        timer.start()
        while True:
            packet = s.recv(4096)
            recvData.extend(packet)
            if(recvData.decode('utf-8').endswith('\r\n') or recvData.decode('utf-8').endswith('\n')):
                break
        message = recvData.decode('utf-8')
        print('RECEIVED MESSAGE: ' + message)
        timer.stop("Received ACK from server")
        
        timer.start()
        if(message.startswith("OK DONE!")):
            if (checkIfProcessRunning("stunnel")):
                lPids = findProcessIdByName("stunnel")
                stunnPid = lPids[0]['pid']
                reconfProc = subprocess.Popen(["sudo", "kill", "-SIGHUP", str(stunnPid)]) # send SIGHUP to stunnel process to force re-configuration
                reconfProc.wait()
                print('Stunnel correctly reconfigured')
                timer.stop("Stunnel reconfigured", True)
            else:
                stunnelThread = Thread(target = startStunnel,)
                stunnelThread.start()
                timer.stop("Stunnel started", True)

